@@ -12,6 +12,10 @@ from pysqream.SQSocket import Client
 import time
 
 
+def _is_null(nullable):
+    return nullable == 1
+
+
 class Cursor:
 
     def __init__(self, conn, cursors):
@@ -34,6 +38,7 @@ class Cursor:
         self.row_size = 0
         self.rows_per_flush = 0
         self.lastrowid = None
+        self.base_connection_closed = False
     
     def get_statement_type(self):
 
@@ -120,6 +125,8 @@ class Cursor:
         self.parsed_rows = []
         self.parsed_row_amount = 0
 
+        if self.ping_loop is not None:
+            _end_ping_loop(self.ping_loop)
         if logger.isEnabledFor(logging.INFO):
             logger.info \
                 (f'Executing statement over connection {self.conn.connection_id} with statement id {self.stmt_id}:\n{stmt}')
@@ -198,17 +205,18 @@ class Cursor:
 
         for type_tup, nullable, tvc in zip(self.col_type_tups, self.col_nul,
                                            self.col_tvc):
-            column = []
-            if nullable:
-                column.append(unsorted_data_columns.pop(0))
-            if tvc:
-                column.append(unsorted_data_columns.pop(0).cast('i'))
+            column = {'nullable': False, 'true_nvarchar': False, 'data_column': False}
 
-            column.append(unsorted_data_columns.pop(0))
+            if nullable:
+                column['nullable'] = unsorted_data_columns.pop(0)
+            if tvc:
+                column['true_nvarchar'] = unsorted_data_columns.pop(0).cast('i')
+
+            column['data_column'] = unsorted_data_columns.pop(0)
             if type_tup[0] not in ('ftVarchar', 'ftBlob', 'ftNumeric'):
-                column[-1] = column[-1].cast(type_to_letter[type_tup[0]])
+                column['data_column'] = column['data_column'].cast(type_to_letter[type_tup[0]])
             else:
-                column[-1] = column[-1].tobytes()
+                column['data_column'] = column['data_column'].tobytes()
             self.data_columns.append(column)
 
         self.unparsed_row_amount = num_rows_fetched
@@ -224,44 +232,24 @@ class Cursor:
             return self.extracted_cols
 
         for idx, raw_col_data in enumerate(self.data_columns):
+
             # Extract data according to column type
             if self.col_tvc[idx]:  # nvarchar
-                nvarc_sizes = raw_col_data[1 if self.col_nul[idx] else 0]
-                col = [
-                    raw_col_data[-1][start:end].decode('utf8')
-                    for (start, end) in lengths_to_pairs(nvarc_sizes)
-                ]
+                col = self._extract_nvarchar(idx , raw_col_data)
+
             elif self.col_type_tups[idx][0] == "ftVarchar":
-                varchar_size = self.col_type_tups[idx][1]
-                col = [
-                    raw_col_data[-1][idx:idx + varchar_size].decode(
-                        self.conn.varchar_enc, "ignore").replace('\x00', '').rstrip()
-                    for idx in range(0, len(raw_col_data[-1]), varchar_size)
-                ]
+                col = self._extract_varchar(idx, raw_col_data)
+
             elif self.col_type_tups[idx][0] == "ftDate":
-                col = [sq_date_to_py_date(d) for d in raw_col_data[-1]]
+                col = self._extract_date(idx, raw_col_data)
+
             elif self.col_type_tups[idx][0] == "ftDateTime":
-                col = [sq_datetime_to_py_datetime(d) for d in raw_col_data[-1]]
+                col = self._extract_datetime(idx, raw_col_data)
+
             elif self.col_type_tups[idx][0] == "ftNumeric":
-                scale = self.col_type_tups[idx][2]
-                col = [
-                    # sq_numeric_to_decimal(bytes_to_bigint(raw_col_data[-1][idx:idx + 16]), scale)
-                    sq_numeric_to_decimal(raw_col_data[-1][idx:idx + 16], scale)
-                    for idx in range(0, len(raw_col_data[-1]), 16)
-                ]
-
+                col = self._extract_numeric(idx, raw_col_data)
             else:
-                col = raw_col_data[-1]
-
-            # Fill Nones if / where needed
-            if self.col_nul[idx]:
-                nulls = raw_col_data[0]  # .tolist()
-                col = [
-                    item if not null else None
-                    for item, null in zip(col, nulls)
-                ]
-            else:
-                pass
+                col = self._extract_datatype(idx, raw_col_data)
 
             self.extracted_cols.append(col)
 
@@ -285,7 +273,10 @@ class Cursor:
     def execute(self, query, params=None):
         """Execute a statement. Parameters are not supported"""
 
-        # self._verify_open()
+        if self.base_connection_closed:
+            self.conn._verify_con_open()
+        else:
+            self.conn._verify_cur_open()
         if params:
 
             log_and_raise(ProgrammingError, "Parametered queries not supported. \
@@ -472,14 +463,85 @@ class Cursor:
             if logger.isEnabledFor(logging.INFO):
                 logger.info(f'Done executing statement {self.stmt_id} over connection {self.conn.connection_id}')
 
+    def _extract_nvarchar(self, idx, raw_col_data):
+        if self.col_nul[idx]:
+            col = [None if (_is_null(n)) else raw_col_data['data_column'][start:end].decode('utf8') for (start, end), n
+                   in
+                   zip(lengths_to_pairs(raw_col_data['true_nvarchar']), raw_col_data['nullable'])]
+        else:
+            col = [
+                raw_col_data['data_column'][start:end].decode('utf8')
+                for (start, end) in lengths_to_pairs(raw_col_data['true_nvarchar'])
+            ]
+        return col
+
+    def _extract_varchar(self, idx, raw_col_data):
+        varchar_size = self.col_type_tups[idx][1]
+        if self.col_nul[idx]:
+            col = []
+            offset = 0
+            for idx in raw_col_data['nullable']:
+                if _is_null(idx):
+                    col.append(None)
+                    offset = offset + varchar_size
+                else:
+                    col.append(raw_col_data['data_column'][offset:offset + varchar_size].decode(self.conn.varchar_enc,
+                                                                                                "ignore").replace(
+                        '\x00', '').rstrip())
+                    offset = offset + varchar_size
+        else:
+            col = [
+                raw_col_data['data_column'][idx:idx + varchar_size].decode(
+                    self.conn.varchar_enc, "ignore").replace('\x00', '').rstrip()
+                for idx in range(0, len(raw_col_data['data_column']), varchar_size)
+            ]
+        return col
+
+    def _extract_date(self, idx, raw_col_data):
+        if self.col_nul[idx]:
+            col = [sq_date_to_py_date(d, is_null=_is_null(n)) for d, n in
+                   zip(raw_col_data['data_column'], raw_col_data['nullable'])]
+        else:
+            col = [sq_date_to_py_date(d) for d in raw_col_data['data_column']]
+        return col
+
+    def _extract_datetime(self, idx, raw_col_data):
+        if self.col_nul[idx]:
+            col = [sq_datetime_to_py_datetime(d, is_null=_is_null(n)) for d, n in
+                   zip(raw_col_data['data_column'], raw_col_data['nullable'])]
+        else:
+            col = [sq_datetime_to_py_datetime(d) for d in raw_col_data['data_column']]
+        return col
+
+    def _extract_numeric(self, idx, raw_col_data):
+        scale = self.col_type_tups[idx][2]
+        if self.col_nul[idx]:
+            col = [
+                sq_numeric_to_decimal(raw_col_data['data_column'][idx:idx + 16], scale, is_null=_is_null(n))
+                for idx, n in zip(range(0, len(raw_col_data['data_column']), 16), raw_col_data['nullable'])
+            ]
+        else:
+            col = [
+                sq_numeric_to_decimal(raw_col_data['data_column'][idx:idx + 16], scale)
+                for idx in range(0, len(raw_col_data['data_column']), 16)
+            ]
+        return col
+
+    def _extract_datatype(self, idx, raw_col_data):
+        if self.col_nul[idx]:
+            col = [None if _is_null(n) else d for d, n in zip(raw_col_data['data_column'], raw_col_data['nullable'])]
+        else:
+            col = raw_col_data['data_column']
+        return col
+
     def close(self, sock=None):
         self.close_stmt()
         sock = sock or self.s
-        self.closed = True
+        self.conn.cur_closed = True
         self.conn.close_connection()
+        self.closed = True
         self.buffer.close()
         _end_ping_loop(self.ping_loop)
-        self.cursors.pop(self.conn.connection_id)
 
     def __enter__(self):
         return self
